@@ -1,38 +1,30 @@
-/* prism2dl.c: user utility for downloading prism2 images
+/* src/prism2/download/prism2dl.c
+*
+* user utility for downloading prism2 images
+*
+* Copyright (C) 1999 AbsoluteValue Systems, Inc.  All Rights Reserved.
 * --------------------------------------------------------------------
 *
-* Linux WLAN 
+* linux-wlan
 *
-*   The contents of this file are subject to the Mozilla Public
-*   License Version 1.1 (the "License"); you may not use this file
-*   except in compliance with the License. You may obtain a copy of
-*   the License at http://www.mozilla.org/MPL/
-*
-*   Software distributed under the License is distributed on an "AS
+*   This software is distributed on an "AS
 *   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *   implied. See the License for the specific language governing
 *   rights and limitations under the License.
 *
-*   The initial developer of the original code is Mark S. Mathews
-*   <mark@absoval.com>.  Portions created by Mark S. Mathews are 
-*   Copyright (C) 1999 AbsoluteValue Software, Inc.  All Rights Reserved.
-*
-*   Alternatively, the contents of this file may be used under the
-*   terms of the GNU Public License version 2 (the "GPL"), in which
-*   case the provisions of the GPL are applicable instead of the
-*   above.  If you wish to allow the use of your version of this file
-*   only under the terms of the GPL and not to allow others to use
-*   your version of this file under the MPL, indicate your decision
-*   by deleting the provisions above and replace them with the notice
-*   and other provisions required by the GPL.  If you do not delete
-*   the provisions above, a recipient may use your version of this
-*   file under either the MPL or the GPL.
-*
+*   This software is NOT distributed under an Open Source license.
+*   If you received it without a direct license from AbsoluteValue 
+*   Systems, Inc. then you are not a licensed user and are violating
+*   AbsoluteValue Systems rights under copyright law.
+*   
 * --------------------------------------------------------------------
 *
-* The initial author may be reached as mark@absoval.com, or 
-* C/O AbsoluteValue Software Inc., P.O. Box 941149, 
-* Maitland, FL, 32794-1149
+* Inquiries regarding this software and AbsoluteValue Systems products
+* and service can be made directly to:
+*
+* AbsoluteValue Systems Inc.
+* info@linux-wlan.com
+* http://www.linux-wlan.com
 *
 * --------------------------------------------------------------------
 *
@@ -40,8 +32,6 @@
 * Intersil Corporation as part of PRISM(R) chipset product development.
 *
 * --------------------------------------------------------------------
-*
-* --------------------------------------------------------------------
 */
 
 /*================================================================*/
@@ -50,21 +40,15 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-
-/* Ugly hack for LinuxPPC R4, don't have time to figure it out right now */
-#if defined(__WLAN_PPC__)
-#undef __GLIBC__
-#endif
-
 #include <unistd.h>
 #include <errno.h>
 #include <ctype.h>
 #include <regex.h>
+#include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
-#include <sys/socket.h>
-#include <asm/spinlock.h>
-#include <asm/byteorder.h>
+#include <endian.h>
+#include <byteswap.h>
 
 /*================================================================*/
 /* Project Includes */
@@ -81,6 +65,19 @@
 #include <wlan/p80211ioctl.h>
 #include <prism2/hfa384x.h>
 
+/* Redefine macros for endianness to avoid using kernel headers */
+#if __BYTE_ORDER == __LITTLE_ENDIAN
+#define hfa384x2host_16(n) (n)
+#define hfa384x2host_32(n) (n)
+#define host2hfa384x_16(n) (n)
+#define host2hfa384x_32(n) (n)
+#else
+#define hfa384x2host_16(n) bswap_16(n)
+#define hfa384x2host_32(n) bswap_32(n)
+#define host2hfa384x_16(n) bswap_16(n)
+#define host2hfa384x_32(n) bswap_32(n)
+#endif
+
 /*================================================================*/
 /* Local Constants */
 
@@ -134,7 +131,7 @@
 
 #define CHUNKS_MAX		100
 
-#define WRITESIZE_MAX		1024
+#define WRITESIZE_MAX		4096
 
 /*================================================================*/
 /* Local Macros */
@@ -209,6 +206,7 @@ int	opt_verbose = 0;	/* -v => boolean, verbose operation */
 int	opt_nowrite = 0;	/* -n => boolean, process all data-but don't download */
 int	opt_debug = 0;		/* -d => boolean, process all data-but don't download */
 int	opt_generate = 0;	/* -g => boolean, if -s or -d output is in pdr and srec file format */
+int	opt_dumpchunks = 0;	/* -D => boolean, dump the imgchunks after plugging and crc */
 
 
 /* IMAGEFILE options */
@@ -231,7 +229,10 @@ UINT8	macaddr[WLAN_ADDR_LEN];		/* -m mac address */
 int	opt_sernum = 0;			/* -S => boolean for cmdline serial # */
 char	sernum[SERNUM_LEN_MAX+1];	/* -S serial # string */
 
-char	opts[] = "svVndgr:f:a:p:m:S:";
+int	opt_pdaloc = 0;			/* -l => pda location */
+					/* -l address */
+
+char	opts[] = "svVndDgr:f:a:p:m:S:l:";
 
 /*----------------------------------------------------------------*/
 /* s-record image processing */
@@ -266,6 +267,10 @@ imgchunk_t	fchunk[CHUNKS_MAX];
 /* PDA, built from [card|newfile]+[addfile1+addfile2...] */
 
 pda_t		pda;
+hfa384x_compident_t nicid;
+hfa384x_caplevel_t  rfid;
+hfa384x_caplevel_t  macid;
+hfa384x_caplevel_t  priid;
 
 const UINT16 crc16tab[256] =
 {
@@ -329,6 +334,9 @@ int	crcimage( imgchunk_t *fchunk, UINT nfchunks,
 int	writeimage(imgchunk_t *fchunk, UINT nfchunks, int isflash);
 void	free_chunks(imgchunk_t *fchunk, UINT* nfchunks);
 void	free_srecs(void);
+void	dumpchunks( imgchunk_t *fchunk, UINT nfchunks);
+
+int     validate_identity(void);
 
 /*================================================================*/
 /* Function Definitions */
@@ -377,7 +385,11 @@ int main ( int argc, char **argv )
 	startaddr = 0;
 
 	nfchunks = 0;
-	memset( fchunk, sizeof(fchunk), 0);
+	memset( fchunk, 0, sizeof(fchunk));
+	memset( &nicid, 0, sizeof(nicid));
+	memset( &rfid, 0, sizeof(rfid));
+	memset( &macid, 0, sizeof(macid));
+	memset( &priid, 0, sizeof(priid));
 
 	/* clear the pda and add an initial END record */
 	memset(&pda, 0, sizeof(pda));
@@ -414,6 +426,10 @@ int main ( int argc, char **argv )
 			opt_debug = 1;
 			opt_verbose = 1;
 			break;
+		case 'D':
+			/* dump chunks */
+			opt_dumpchunks = 1;
+			break;
 		case 's':
 			/* Show status */
 			opt_status = 1;
@@ -459,6 +475,9 @@ int main ( int argc, char **argv )
 			opt_newpda = 1;
 			strncpy(newpdafname, optarg, FILENAME_MAX);
 			break;
+		case 'l':
+			opt_pdaloc = strtoul(optarg, NULL, 0);
+			break;
 		case 'm':
 			/* cmdline MAC address */
 			if ( opt_macaddr ) {
@@ -529,54 +548,92 @@ int main ( int argc, char **argv )
 		}
 	}
 
-	/* If the MAC address is specified it should overwrite the
-           current value. */
-	if ( opt_macaddr ) {
-	  const unsigned nwords = WLAN_ADDR_LEN/sizeof(UINT16) + 2;
-	  UINT16 pdwords[nwords];
-	  int i; /* index into byte array of macaddr */
-	  int j; /* index into pdr words */
+	/* read the card's PRI-SUP */
+	if (!opt_debug) {
+		p80211msg_dot11req_mibget_t getmsg;
+		p80211itemd_t *item;
+		UINT32        *data;
 
-	  /* create mac address PDR
-	     word[0] : PDR length
-	     word[1] : MAC address PDR code (0x101)
-	     word[2-4] : MAC address (WLAN_ADDR_LEN bytes - usually 6) 
-	  */
-	  pdwords[0] = host2hfa384x_16(0x0004);
-	  pdwords[1] = host2hfa384x_16(0x0101);
-	  for (i = 0, j = 2; i < WLAN_ADDR_LEN; i += 2, j++) {
-	    pdwords[j] = host2hfa384x_16((UINT16) macaddr[i+1] << 8 | (UINT16) macaddr[i]);
-	  }
-	  /* merge the value into the PDA, so that it will eventually
-             be written. */
-	  merge_pda(&pda, pdwords, nwords);
-	  pda_changed = 1;
+		memset(&getmsg, 0, sizeof(getmsg));
+		getmsg.msgcode = DIDmsg_dot11req_mibget;
+		getmsg.msglen = sizeof(getmsg);
+		strcpy(getmsg.devname, devname);
+
+		getmsg.mibattribute.did = DIDmsg_dot11req_mibget_mibattribute;
+		getmsg.mibattribute.status = P80211ENUM_msgitem_status_data_ok;
+		getmsg.resultcode.did = DIDmsg_dot11req_mibget_resultcode;
+		getmsg.resultcode.status = P80211ENUM_msgitem_status_no_value;
+
+		item = (p80211itemd_t *) getmsg.mibattribute.data;
+		item->did = DIDmib_p2_p2NIC_p2PRISupRange;
+		item->status = P80211ENUM_msgitem_status_no_value;
+		
+		data = (UINT32*) item->data;
+
+		do_ioctl((p80211msg_t*)&getmsg);
+		if (getmsg.resultcode.data != P80211ENUM_resultcode_success) {
+			printf("Couldn't fetch PRI-SUP info\n");
+		}
+
+		/* Already in host order */
+		priid.role = *data++;
+		priid.id = *data++;
+		priid.variant = *data++;
+		priid.bottom = *data++;
+		priid.top = *data++;
+	}
+
+	/* If the MAC address is specified it should overwrite the
+         * current value. 
+	 */
+	if ( opt_macaddr ) {
+		const unsigned nwords = WLAN_ADDR_LEN/sizeof(UINT16) + 2;
+		UINT16 pdwords[nwords];
+		int i; /* index into byte array of macaddr */
+		int j; /* index into pdr words */
+
+		/* create mac address PDR
+		word[0] : PDR length
+		word[1] : MAC address PDR code (0x101)
+		word[2-4] : MAC address (WLAN_ADDR_LEN bytes - usually 6) 
+		*/
+		pdwords[0] = host2hfa384x_16(0x0004);
+		pdwords[1] = host2hfa384x_16(0x0101);
+		for (i = 0, j = 2; i < WLAN_ADDR_LEN; i += 2, j++) {
+			pdwords[j] = host2hfa384x_16((UINT16) macaddr[i+1] << 8 | (UINT16) macaddr[i]);
+		}
+		/* merge the value into the PDA, so that it will eventually
+		 * be written. 
+		 */
+		merge_pda(&pda, pdwords, nwords);
+		pda_changed = 1;
 	}
 
 	/* If a serial number is specified add it to the PDA. */
 	if ( opt_sernum ) {
-	  const unsigned nwords = SERNUM_LEN_MAX/sizeof(UINT16) + 2;
-	  UINT16 pdwords[nwords];
-	  int i; /* index into byte array of serial numbers bytes */
-	  int j; /* index into pdr words */
+		const unsigned nwords = SERNUM_LEN_MAX/sizeof(UINT16) + 2;
+		UINT16 pdwords[nwords];
+		int i; /* index into byte array of serial numbers bytes */
+		int j; /* index into pdr words */
 
-	  /* create mac address PDR
-	     word[0] : PDR length
-	     word[1] : Serial number PDR code (0x0003)
-	     word[2-7] : Serial number (12 bytes)
-	  */
-	  pdwords[0] = host2hfa384x_16(0x0007);
-	  pdwords[1] = host2hfa384x_16(0x0003);
-	  for (i = 0, j = 2; i < SERNUM_LEN_MAX; i += 2, j++) {
-	    pdwords[j] = host2hfa384x_16((UINT16) sernum[i+1] << 8 | (UINT16) sernum[i]);
-	  }
-	  /* merge the value into the PDA, so that it will eventually
-             be written. */
-	  merge_pda(&pda, pdwords, nwords);
-	  pda_changed = 1;
+		/* create mac address PDR
+		word[0] : PDR length
+		word[1] : Serial number PDR code (0x0003)
+		word[2-7] : Serial number (12 bytes)
+		*/
+		pdwords[0] = host2hfa384x_16(0x0007);
+		pdwords[1] = host2hfa384x_16(0x0003);
+		for (i = 0, j = 2; i < SERNUM_LEN_MAX; i += 2, j++) {
+			pdwords[j] = host2hfa384x_16((UINT16) sernum[i+1] << 8 | (UINT16) sernum[i]);
+		}
+		/* merge the value into the PDA, so that it will eventually
+		 * be written. 
+		 */
+		merge_pda(&pda, pdwords, nwords);
+		pda_changed = 1;
 	}
 	  
-	if ( opt_status ) { /* print pda */
+	if ( opt_status || opt_generate ) { /* print pda */
 		print_all_pdrs(&pda);
 	}
 
@@ -591,19 +648,32 @@ int main ( int argc, char **argv )
 		mkpda_crc(&pda);
 	}
 
-	if ( opt_status && pda_changed ) {
+	if ( (opt_status || opt_generate ) && pda_changed ) {
 		printf("PDA after crc calc and/or \"-a\" file "
 			"generated modifications:\n");
 		print_all_pdrs(&pda);
 		/* Read and print the CIS? */
 		exit(0);
-	} else if  ( opt_status ) { /* We're done */
+	} else if  ( opt_status || opt_generate ) { /* We're done */
 		exit(0);  
 	}
 
+	if ( opt_ramloadcnt && !opt_flashloadcnt ) {
+		if (pda_changed) {
+		printf("Warning: RAM load only, PDA changes will NOT be written to flash.\n");
+		}
+		goto skip_pda_write;
+	}
+
 	if ( !opt_debug && !opt_nowrite && pda_changed) { /* write the PDA */
+		if ( opt_pdaloc == 0 ) {
+			fprintf(stderr, APPNAME ": error, you must specifify a pda location\n");
+			usage();
+			exit(1);
+		}
 		pda_write(&pda);
 	}
+skip_pda_write:
 
 	/*-----------------------------------------------------*/
 	/* Read the flash files */
@@ -623,7 +693,18 @@ int main ( int argc, char **argv )
 				sizeof(s3datarec_t),
 				s3datarec_compare);
 
-			/* TODO: Validate the identity and compat ranges */
+			result = validate_identity();
+
+			if (startaddr != 0x00000000) {
+				fprintf(stderr, APPNAME ": Can't flash a RAM download image!\n");
+				exit(1);
+			}
+
+			if ( result ) {
+				fprintf(stderr, APPNAME ": Incompatible firmware image.\n");
+				exit(1);
+			}
+
 			/* Make the image chunks */
 			result = mkimage(fchunk, &nfchunks);
 
@@ -679,7 +760,18 @@ int main ( int argc, char **argv )
 				sizeof(s3datarec_t),
 				s3datarec_compare);
 
-			/* TODO: Validate the identity and compat ranges */
+			result = validate_identity();
+
+			if ( result && !opt_newpda ) {
+				fprintf(stderr, APPNAME ": Incompatible firmware image.\n");
+				exit(1);
+			}
+
+			if (startaddr == 0x00000000) {
+				fprintf(stderr, APPNAME ": Can't RAM download a Flash image!\n");
+				exit(1);
+			}
+
 			/* Make the image chunks */
 			result = mkimage(fchunk, &nfchunks);
 
@@ -702,6 +794,11 @@ int main ( int argc, char **argv )
 				exit(1);
 			}
 
+			if ( opt_dumpchunks ) {
+				/* Dump the contents of the image chunks */
+				dumpchunks(fchunk, nfchunks);
+			}
+
 			/* Write the image */
 			if ( opt_nowrite ) continue;
 			result = writeimage(fchunk, nfchunks, 0);
@@ -749,7 +846,7 @@ int crcimage(imgchunk_t *fchunk, UINT nfchunks, s3crcrec_t *s3crc, UINT ns3crc)
 	int	c;
 	UINT32	crcstart;
 	UINT32	crcend;
-	UINT32	cstart;
+	UINT32	cstart = 0;
 	UINT32	cend;
 	UINT8	*dest;
 	UINT32	chunkoff;
@@ -825,7 +922,7 @@ int do_ioctl( p80211msg_t *msg )
 		perror(APPNAME);
 	} else {
 		req.len = msg->msglen;
-		req.data = msg;
+		req.data = (caddr_t) msg;
 		strcpy( req.name, msg->devname);
 		req.result = 0;
 
@@ -861,7 +958,7 @@ void free_chunks(imgchunk_t *fchunk, UINT* nfchunks)
 		}
 	}
 	*nfchunks = 0;
-	memset( fchunk, sizeof(fchunk), 0);
+	memset( fchunk, 0, sizeof(fchunk));
 
 }
 
@@ -918,7 +1015,7 @@ int mkimage(imgchunk_t *clist, UINT *ccnt)
 	UINT32		nextaddr = 0;
 	UINT32		s3start;
 	UINT32		s3end;
-	UINT32		cstart;
+	UINT32		cstart = 0;
 	UINT32		cend;
 	UINT32		coffset;
 	
@@ -1030,7 +1127,7 @@ int mkpda_crc( pda_t *pda)
 	pda->rec[pda->nrec - 1]->data.end_of_pda.crc = host2hfa384x_16(crc);
 
 	if (opt_debug) {
-		printf(__FUNCTION__": pdacrc=0x%04x\n", crc);
+		printf("%s: pdacrc=0x%04x\n", __FUNCTION__, crc);
 	}
 
 	return result;
@@ -1041,11 +1138,6 @@ int mkpda_crc( pda_t *pda)
 * mkpdrlist
 *
 * Reads a raw PDA and builds an array of pdrec_t structures.
-* NOTE: the len and code fields of the pdrrec_t's in the array
-*       will contain their values in HOST byte order following the
-*       execution of this function.  The data field is NOT converted.
-*       If you use one of the data field elements don't forget to
-*       convert it first.
 *
 * Arguments:
 *	pda	buffer containing raw PDA bytes
@@ -1065,14 +1157,46 @@ int mkpdrlist( pda_t *pda)
 	pda->nrec = 0;
 	curroff = 0;
 	while ( curroff < (HFA384x_PDA_LEN_MAX / 2) &&
-		hfa384x2host_16(pda16[curroff + 1]) != HFA384x_PDR_END_OF_PDA ) {
+		hfa384x2host_16(pda16[curroff + 1]) !=
+		HFA384x_PDR_END_OF_PDA ) {
 		pda->rec[pda->nrec] = (hfa384x_pdrec_t*)&(pda16[curroff]);
+
+		if (hfa384x2host_16(pda->rec[pda->nrec]->code) ==
+				HFA384x_PDR_NICID) {
+			memcpy(&nicid, &pda->rec[pda->nrec]->data.nicid,
+			       sizeof(nicid));
+			nicid.id = hfa384x2host_16(nicid.id);
+			nicid.variant = hfa384x2host_16(nicid.variant);
+			nicid.major = hfa384x2host_16(nicid.major);
+			nicid.minor = hfa384x2host_16(nicid.minor);
+		}
+		if (hfa384x2host_16(pda->rec[pda->nrec]->code) ==
+				HFA384x_PDR_MFISUPRANGE) {
+			memcpy(&rfid, &pda->rec[pda->nrec]->data.mfisuprange,
+			       sizeof(rfid));
+			rfid.id = hfa384x2host_16(rfid.id);
+			rfid.variant = hfa384x2host_16(rfid.variant);
+			rfid.bottom = hfa384x2host_16(rfid.bottom);
+			rfid.top = hfa384x2host_16(rfid.top);
+		}
+		if (hfa384x2host_16(pda->rec[pda->nrec]->code) ==
+				HFA384x_PDR_CFISUPRANGE) {
+			memcpy(&macid, &pda->rec[pda->nrec]->data.cfisuprange,
+			       sizeof(macid));
+			macid.id = hfa384x2host_16(macid.id);
+			macid.variant = hfa384x2host_16(macid.variant);
+			macid.bottom = hfa384x2host_16(macid.bottom);
+			macid.top = hfa384x2host_16(macid.top);
+		}
+
 		(pda->nrec)++;
 		curroff += hfa384x2host_16(pda16[curroff]) + 1;
+
 	}
 	if ( curroff >= (HFA384x_PDA_LEN_MAX / 2) ) {
 		fprintf(stderr, APPNAME
-			": no end record found or invalid lengths in PDR data, exiting.\n");
+			": no end record found or invalid lengths in "
+			"PDR data, exiting. %x %d\n", curroff, pda->nrec);
 		exit(1);
 	}
 	if (hfa384x2host_16(pda16[curroff + 1]) == HFA384x_PDR_END_OF_PDA ) {
@@ -1130,7 +1254,7 @@ int pda_write(pda_t *pda)
 	writemsg.resultcode.status =	P80211ENUM_msgitem_status_no_value;
 	writemsg.addr.len =		sizeof(UINT32);
 	writemsg.len.len =		sizeof(UINT32);
-	writemsg.data.len =		1024;
+	writemsg.data.len =		WRITESIZE_MAX;
 	writemsg.resultcode.len =	sizeof(UINT32);
 
 	/* Send flashdl_state(enable) */
@@ -1154,7 +1278,7 @@ int pda_write(pda_t *pda)
 	}
 
 	/* Send flashdl_write(pda) */
-	writemsg.addr.data = HFA384x_PDA_BASE;
+	writemsg.addr.data = opt_pdaloc;
 	writemsg.len.data = (((UINT8*)(pda->rec[pda->nrec - 1])) - pda->buf + 6);
 	memcpy(writemsg.data.data, pda->buf, writemsg.len.data);
 	if (opt_verbose) {
@@ -1229,7 +1353,7 @@ int plugimage( imgchunk_t *fchunk, UINT nfchunks,
 	int	c;	/* chunk index */
 	UINT32	pstart;
 	UINT32	pend;
-	UINT32	cstart;
+	UINT32	cstart = 0;
 	UINT32	cend;
 	UINT32	chunkoff;
 	UINT8	*src;
@@ -1250,11 +1374,22 @@ int plugimage( imgchunk_t *fchunk, UINT nfchunks,
 		}
 		if ( j >= pda->nrec && j != -1 ) { /*  if no matching PDR, fail */
 			fprintf(stderr, APPNAME
-				": error: Failed to find PDR for "
+				": warning: Failed to find PDR for "
 				"plugrec 0x%04lx.\n",
 				s3plug[i].itemcode);
+			continue; /* and move on to the next PDR */
+
+#if 0
+			/* MSM: They swear that unless it's the MAC address,
+			 * the serial number, or the TX calibration records,
+			 * then there's reasonable defaults in the f/w
+			 * image.  Therefore, missing PDRs in the card
+			 * should only be a warning, not fatal.
+			 * TODO: add fatals for the PDRs mentioned above.
+			 */
 			result = 1;
-			continue;
+			continue; 
+#endif
 		}
 
 		/* Validate plug len against PDR len */
@@ -1282,11 +1417,6 @@ int plugimage( imgchunk_t *fchunk, UINT nfchunks,
 			result = 1;
 			continue;
 		}
-		/* An error has occurred, don't bother doing the actual plug */
-		/* But we want to continue so we detect all the errors */
-		if ( result != 0 ) {
-			continue;
-		}
 
 		/* Plug data */
 		chunkoff = pstart - cstart;
@@ -1357,7 +1487,7 @@ void print_all_pdrs(pda_t *pda)
 		datap = (UINT16*)&(pda->rec[i]->data.end_of_pda);
 		nwords = hfa384x2host_16(pda->rec[i]->len) - 1;
 		for ( j = 0; j < nwords; j++) {
-			printf("%04x ", datap[j] );
+			printf("%04x ", hfa384x2host_16(datap[j]) );
 			if ( (j % 8) == 7 && j < nwords - 1 ) {
 				printf("\n                         ");
 			}
@@ -1371,14 +1501,13 @@ void print_all_pdrs(pda_t *pda)
 		end = (((UINT8*)pda->rec[pda->nrec - 1]) - pda->buf + 6) / 2;
 		printf("  ");
 		for ( i = 0; i < end; i++ ) {
-			printf("%04x ", datap[i]);
+			printf("%04x ", hfa384x2host_16(datap[i]) );
 			if ( (i % 16) == 15 ) {
 				printf("\n  ");
 			}
 		}
 		printf("\n");
 	}
-
 }
 
 
@@ -1420,7 +1549,11 @@ int read_cardpda(pda_t *pda, char *dev)
 	} else if ( msg.resultcode.data == P80211ENUM_resultcode_success ) {
 		memcpy(pda->buf, msg.pda.data, HFA384x_PDA_LEN_MAX);
 		result = mkpdrlist(pda);
+	} else {
+		/* resultcode must've been something other than success */
+		result = -1;
 	}
+
 	return result;
 }
 
@@ -1667,11 +1800,15 @@ int read_srecfile(char *fname)
 	
 
 	printf("Reading S-record file %s...\n", fname);
-	f = fopen(fname, "r");
-	if ( f == NULL ) {
-		result=errno;
-		perror(APPNAME);
-		return result;
+	if ( strcmp("stdin", fname) == 0 ) {
+		f = stdin;
+	} else {
+		f = fopen(fname, "r");
+		if ( f == NULL ) {
+			result=errno;
+			perror(APPNAME);
+			return result;
+		}
 	}
 
 	while ( fgets(buf, sizeof(buf), f) != NULL ) {
@@ -1690,7 +1827,8 @@ int read_srecfile(char *fname)
 					line, 
 					startaddr);
 			}
-			break;
+//			break;
+			continue;
 		} else if ( buf[1] != '3') {
 			fprintf(stderr,APPNAME":%s:%d warning: Unknown S-record detected.\n", fname, line);
 			fclose(f);
@@ -1714,17 +1852,17 @@ int read_srecfile(char *fname)
 			memcpy(tmpbuf, buf+S3PLUG_ITEMCODE_TXTOFFSET, S3PLUG_ITEMCODE_TXTLEN);
 			tmpbuf[S3PLUG_ITEMCODE_TXTLEN] = '\0';
 			s3plug[ns3plug].itemcode = strtoul(tmpbuf,NULL,16);
-			s3plug[ns3plug].itemcode =__swab32(s3plug[ns3plug].itemcode);
+			s3plug[ns3plug].itemcode = bswap_32(s3plug[ns3plug].itemcode);
 
 			memcpy(tmpbuf, buf+S3PLUG_ADDR_TXTOFFSET, S3PLUG_ADDR_TXTLEN);
 			tmpbuf[S3PLUG_ADDR_TXTLEN] = '\0';
 			s3plug[ns3plug].addr = strtoul(tmpbuf,NULL,16);
-			s3plug[ns3plug].addr = __swab32(s3plug[ns3plug].addr);
+			s3plug[ns3plug].addr = bswap_32(s3plug[ns3plug].addr);
 
 			memcpy(tmpbuf, buf+S3PLUG_LEN_TXTOFFSET, S3PLUG_LEN_TXTLEN);
 			tmpbuf[S3PLUG_LEN_TXTLEN] = '\0';
 			s3plug[ns3plug].len = strtoul(tmpbuf,NULL,16);
-			s3plug[ns3plug].len = __swab32(s3plug[ns3plug].len);
+			s3plug[ns3plug].len = bswap_32(s3plug[ns3plug].len);
 
 			if (opt_verbose) {
 				printf( "  S3 plugrec, line=%d "
@@ -1741,17 +1879,17 @@ int read_srecfile(char *fname)
 			memcpy(tmpbuf, buf+S3CRC_ADDR_TXTOFFSET, S3CRC_ADDR_TXTLEN);
 			tmpbuf[S3CRC_ADDR_TXTLEN] = '\0';
 			s3crc[ns3crc].addr = strtoul(tmpbuf,NULL,16);
-			s3crc[ns3crc].addr =__swab32(s3crc[ns3crc].addr);
+			s3crc[ns3crc].addr = bswap_32(s3crc[ns3crc].addr);
 
 			memcpy(tmpbuf, buf+S3CRC_LEN_TXTOFFSET, S3CRC_LEN_TXTLEN);
 			tmpbuf[S3CRC_LEN_TXTLEN] = '\0';
 			s3crc[ns3crc].len = strtoul(tmpbuf,NULL,16);
-			s3crc[ns3crc].len =__swab32(s3crc[ns3crc].len);
+			s3crc[ns3crc].len = bswap_32(s3crc[ns3crc].len);
 
 			memcpy(tmpbuf, buf+S3CRC_DOWRITE_TXTOFFSET, S3CRC_DOWRITE_TXTLEN);
 			tmpbuf[S3CRC_DOWRITE_TXTLEN] = '\0';
 			s3crc[ns3crc].dowrite = strtoul(tmpbuf,NULL,16);
-			s3crc[ns3crc].dowrite =__swab32(s3crc[ns3crc].dowrite);
+			s3crc[ns3crc].dowrite = bswap_32(s3crc[ns3crc].dowrite);
 
 			if (opt_verbose) {
 				printf( "  S3 crcrec, line=%d "
@@ -1767,19 +1905,19 @@ int read_srecfile(char *fname)
 			memcpy(tmpbuf, buf+S3INFO_LEN_TXTOFFSET, S3INFO_LEN_TXTLEN);
 			tmpbuf[S3INFO_LEN_TXTLEN] = '\0';
 			s3info[ns3info].len = strtoul(tmpbuf,NULL,16);
-			s3info[ns3info].len =__swab16(s3info[ns3info].len);
+			s3info[ns3info].len = bswap_16(s3info[ns3info].len);
 
 			memcpy(tmpbuf, buf+S3INFO_TYPE_TXTOFFSET, S3INFO_TYPE_TXTLEN);
 			tmpbuf[S3INFO_TYPE_TXTLEN] = '\0';
 			s3info[ns3info].type = strtoul(tmpbuf,NULL,16);
-			s3info[ns3info].type =__swab16(s3info[ns3info].type);
+			s3info[ns3info].type = bswap_16(s3info[ns3info].type);
 
 			tmpinfo = (UINT16*)&(s3info[ns3info].info.version);
 			for (i = 0; i < s3info[ns3info].len - 1; i++) {
 				memcpy( tmpbuf, buf+S3INFO_DATA_TXTOFFSET+(i*4), 4);
 				tmpbuf[4] = '\0';
 				tmpinfo[i] = strtoul(tmpbuf,NULL,16);
-				tmpinfo[i] = __swab16(tmpinfo[i]);
+				tmpinfo[i] = bswap_16(tmpinfo[i]);
 			}
 			if (opt_verbose) {
 				printf( "  S3 inforec, line=%d "
@@ -1792,8 +1930,8 @@ int read_srecfile(char *fname)
 					printf("%04x ", tmpinfo[i]);
 				}
 				printf("\n");
-
 			}
+
 			ns3info++;
 			break;
 		default:	/* Data record */
@@ -1910,20 +2048,27 @@ void usage(void)
 "\n"
 "       IMAGEFILE OPTIONS:\n"
 "       -r <file>   (ram)      Load SREC file to card RAM.  This option may\n"
-"                              be specified multiple times.\n"
+"                              be specified multiple times.  If the value \n"
+"                              is \"stdin\", the file will be read from \n"
+"                              stdin and the option may only be specified once.\n"
 "       -f <file>   (flash)    Load SREC file to card FLASH. This option may\n"
-"                              be specified multiple times.\n"
+"                              be specified multiple times.  If the value \n"
+"                              is \"stdin\", the file will be read from \n"
+"                              stdin and the option may only be specified once.\n"
 "\n"
 "       PDA OPTIONS:\n"
 "       -a <file>   (addpdr)   Add the PDRs from file to the PDA from card. This\n"
 "                              option may be specified multiple times.\n"
 "       -p <file>   (pda)      Replace the card PDA with the contents of file.\n"
-"       -m <addr>   (macaddr)  Overwrite the MAC address PDR with the given \n"
+"       -m <haddr>  (macaddr)  Overwrite the MAC address PDR with the given \n"
 "                              value.  <addr> ::= xx:xx:xx:xx:xx:xx, where \n"
 "                              xx is a two digit hex number.\n"
 "       -S <str>    (Sernum)   Overwrite the serial number PDR with the given\n"
 "                              string.  String must be <= 12 characters, any\n"
 "                              extra will be truncated.\n"
+"       -l <addr>   (pdaloc)   PDA location in card memory.  Commonly values:\n"
+"                                 HFA3841 ==> 0x003f0000\n"
+"                                 HFA3842 ==> 0x007f0000\n"
 "\n"
 "  argument:\n"
 "  	devname    Linux device name (e.g. eth0, wlan0)\n"
@@ -2016,7 +2161,7 @@ int writeimage(imgchunk_t *fchunk, UINT nfchunks, int isflash)
 	fwritemsg.resultcode.status =	P80211ENUM_msgitem_status_no_value;
 	fwritemsg.addr.len =		sizeof(UINT32);
 	fwritemsg.len.len =		sizeof(UINT32);
-	fwritemsg.data.len =		1024;
+	fwritemsg.data.len =		WRITESIZE_MAX;
 	fwritemsg.resultcode.len =	sizeof(UINT32);
 
 	memset(&rstatemsg, 0, sizeof(rstatemsg));
@@ -2047,7 +2192,7 @@ int writeimage(imgchunk_t *fchunk, UINT nfchunks, int isflash)
 	rwritemsg.resultcode.status =	P80211ENUM_msgitem_status_no_value;
 	rwritemsg.addr.len =		sizeof(UINT32);
 	rwritemsg.len.len =		sizeof(UINT32);
-	rwritemsg.data.len =		1024;
+	rwritemsg.data.len =		WRITESIZE_MAX;
 	rwritemsg.resultcode.len =	sizeof(UINT32);
 
 	/* Send xxx_state(enable) */
@@ -2183,4 +2328,139 @@ fclose(fp);
 }
 
 
+/*----------------------------------------------------------------
+* dumpchunks
+*
+* Dumps the chunk info for each chunk of the current set to stdout.
+* 
+*
+* Arguments:
+*	fchunk		Array of image chunks
+*	nfchunks	Number of image chunks
+*
+* Returns: 
+*	0	success
+*	~0	failure
+----------------------------------------------------------------*/
+void dumpchunks( imgchunk_t *fchunk, UINT nfchunks)
+{
+	int	i;
+	int	j;
 
+	for ( i = 0; i < nfchunks; i++) {
+		printf("\nChunk %d: addr=0x%08lx  len=%ld  crc=%04hx\n",
+			i, fchunk[i].addr, fchunk[i].len, fchunk[i].crc);
+		printf("%08lx: ", fchunk[i].addr);
+		for(j = 0; j < fchunk[i].len; j++) {
+			printf("%02x ", fchunk[i].data[j]);
+			if (((j+1) % 16) == 0) {
+				printf("\n%08lx: ", fchunk[i].addr + j + 1);
+			}
+		}
+	}
+	printf("\n");
+}
+
+int validate_identity(void) 
+{
+	int i;
+	int result = 1;
+
+	printf("NIC ID: %#x v%d.%d.%d\n", 
+	       nicid.id,
+	       nicid.major,
+	       nicid.minor,
+	       nicid.variant);
+	printf("MFI ID: %#x v%d %d->%d\n", 
+	       rfid.id,
+	       rfid.variant,
+	       rfid.bottom,
+	       rfid.top);
+	printf("CFI ID: %#x v%d %d->%d\n", 
+	       macid.id,
+	       macid.variant,
+	       macid.bottom,
+	       macid.top);
+	printf("PRI ID: %#x v%d %d->%d\n",
+	       priid.id,
+	       priid.variant,
+	       priid.bottom,
+	       priid.top);
+
+	for (i = 0 ; i < ns3info ; i ++) {
+		switch (s3info[i].type) {
+		case 1:
+			printf("Version:  ID %#x %d.%d.%d\n", 
+			       s3info[i].info.version.id,
+			       s3info[i].info.version.major,
+				       s3info[i].info.version.minor,
+			       s3info[i].info.version.variant);
+			break;
+		case 2:
+			printf("Compat: Role %#x Id %#x v%d %d->%d\n",
+			       s3info[i].info.compat.role,
+			       s3info[i].info.compat.id,
+			       s3info[i].info.compat.variant,
+			       s3info[i].info.compat.bottom,
+			       s3info[i].info.compat.top);
+
+			/* MAC compat range */
+			if ((s3info[i].info.compat.role == 1) &&
+			    (s3info[i].info.compat.id == 2)) {
+				if (s3info[i].info.compat.variant != 
+				    macid.variant) {
+					result = 2;
+				}	
+			}
+		
+			/* PRI compat range */
+			if ((s3info[i].info.compat.role == 1) &&
+			    (s3info[i].info.compat.id == 3)) {
+				if ((s3info[i].info.compat.bottom > priid.top) ||
+				    (s3info[i].info.compat.top < priid.bottom)){
+					result = 3;
+				}					
+			}
+			/* SEC compat range */
+			if ((s3info[i].info.compat.role == 1) &&
+			    (s3info[i].info.compat.id == 4)) {
+				
+			}
+
+			break;
+		case 3:
+				printf("Seq: %#x\n", s3info[i].info.buildseq);
+				
+				break;
+		case 4:
+			printf("Platform:  ID %#x %d.%d.%d\n", 
+			       s3info[i].info.version.id,
+			       s3info[i].info.version.major,
+			       s3info[i].info.version.minor,
+			       s3info[i].info.version.variant);
+
+			if (nicid.id != s3info[i].info.version.id)
+				continue;
+			if (nicid.major != s3info[i].info.version.major)
+				continue;
+			if (nicid.minor != s3info[i].info.version.minor)
+				continue;
+			if ((nicid.variant != s3info[i].info.version.variant) &&
+			    (nicid.id != 0x8008))
+				continue;
+
+			if (result != 2)
+				result = 0;
+			break;
+		case 0x8001:
+			printf("name inforec len %d\n", s3info[i].len);
+			
+			break;
+		default:
+			printf("Unknown inforec type %d\n", s3info[i].type);
+		}
+	}
+	// walk through 
+
+	return result;
+}
